
/* $Id: csimple.c,v 3.13 1999/11/11 04:51:34 kohl Exp $ */


/*
 *                      CUMULVS version 1.0:
 *      Computational Steering and Interactive Visualization
 *                   in Distributed Applications
 *           Oak Ridge National Laboratory, Oak Ridge TN.
 *      Authors:  James Arthur Kohl and Philip M. Papadopoulos
 *                   (C) 1996 All Rights Reserved
 *
 *                              NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Neither the Institution, Oak Ridge National Laboratory, nor the
 * Authors make any representations about the suitability of this
 * software for any purpose.  This software is provided ``as is''
 * without express or implied warranty.
 *
 * CUMULVS was funded by the U.S. Department of Energy.
 */


/*
	csimple.c - A small test program to test the communication
	setup with CUMULVS.
*/


#include <stdio.h>
#include <math.h>
#include "pvm3.h"
#include "stv.h"


/* ---  Adjustable Parameters  --- */


/* The number of processes used is ROWS*COLUMNS */

#define ROWS 2
#define COLUMNS 2


/* The size of the computational region is 
	(ROWS * I_DIM) X (COLUMNS* J_DIM) X (K_DIM)
*/

#define IDIM 30
#define JDIM 30
#define KDIM 120


/* --- End Adjustable Parameters --- */

#define CSIMPLEGRP "csimple"
#define STVTAG 100
#define RINGMSG 1001
#define NTASKS (ROWS*COLUMNS)
#define DEPTH KDIM
#define NFIELDS 1
#define NPARAMS 2


/* --- Square of x Macro --- */

#define SQR(x) ((x) * (x))


/* --- Other Options --- */

/* #define INTERMEDIATE_SEND */

/* #define VERBOSE    stvTrue */
#define VERBOSE    stvFalse


void dump_params_changed();


/* --- Main Program --- */

void
main()
{

/* --  Variable declarations  -- */

	void *paramPtr;	

	char *paramName;

	int axisType[3];
	int bufid;
	int cc, centeri, centerj, centerk;
	int declare[3], decompId[NFIELDS], dest;
	int fieldId[NFIELDS];
	int glb[3], globali, globalj, globalk, gub[3];
	int i, info;
	int j;
	int k;
	int mycol, myinum, myrow, msrc, mytid;
	int nspawn;
	int onum, otid;
	int paramIds[NPARAMS], prank, pshape[2], paddr[2], paramType;
	int tids[NTASKS];
	int sequence, src;

	float density[IDIM][JDIM][KDIM];
	float dist;
	float radius, rate; 
	float vradius, vrate;
	float value;

/* --  PVM Initialization  -- */

	mytid = pvm_mytid();
	myinum = pvm_joingroup( CSIMPLEGRP );

/*  --- Should I spawn other tasks? */

	if ( myinum == 0 )
	{ 
		if ( NTASKS - 1 > 0 ) 
		{
			pvm_catchout( stdout );

			nspawn = pvm_spawn( "csimple", (char **) 0,
					PvmTaskDefault,
					/* PvmTaskDebug, */
					" ", NTASKS - 1, tids );

			if ( nspawn < NTASKS - 1 )
			{
				printf( "Spawn failed with nspawn = %d\n", nspawn ); 

				printf( "Killing orphaned tasks...\n" );

				for ( i=0 ; i < nspawn ; i++ ) 
					pvm_kill( tids[i] );

				printf( "Failed spawns returned codes:\n" );

				for ( i=nspawn ; i <  NTASKS - 1 ; i++ ) 
					printf( "Info was: %d\n", tids[i] );	

				pvm_exit();

				exit( -1 );
			}
		}
	}

/*  --- Make the group static -- this is also a barrier  */

	pvm_freezegroup( CSIMPLEGRP, NTASKS );

/*  --- Determine src and destination for ring message */

	if ( (dest = myinum + 1) == NTASKS )
		dest = 0;

	dest = pvm_gettid( CSIMPLEGRP, dest ); 

	if ( (src = myinum - 1) < 0 ) 
		src = NTASKS - 1;

	src = pvm_gettid( CSIMPLEGRP, src );

	/* notify me if anyone fails */
	for ( i=0 ; i < NTASKS ; i++ )
	{
		cc = pvm_gettid( CSIMPLEGRP, i );

		pvm_notify( PvmTaskExit, RINGMSG, 1, &cc );
	}

/* --- Initialize CUMULVS steering and visualization */

	stv_setopt( stvOptNull, stvOptVerbose,
		VERBOSE ? STV_DEBUG_ALL : 0 );

	/* stv_setopt( stvOptNull, stvOptAppLock, stvTrue ); */

	stv_init( CSIMPLEGRP, STVTAG, NTASKS, myinum );

/* --- Record the size of the computational region */

	myrow = myinum / COLUMNS;
	mycol = myinum %  COLUMNS;

	glb[0] = 1; glb[1] = 1; glb[2] = 1;

	gub[0] = ROWS * IDIM;
	gub[1] = COLUMNS * JDIM;
	gub[2] = KDIM;
		
/* --- Initialize the parameters */

	radius = 0;
	rate = .2;
	vrate = 1.0;
	sequence = 0;
	
	centeri = ( IDIM * ROWS + 1 ) / 2;
	centerj = ( JDIM * COLUMNS + 1 ) / 2;
	centerk = ( KDIM + 1 ) / 2;

	if ( myinum == 0 ) 
		printf( "center: %d %d %d \n", centeri, centerj, centerk );

/* --- Define the steering parameters */

	paramIds[0] = stv_paramDefine( "Radius Rate", (void *)&rate,
			stvFloat, stvVisOnly );

	paramIds[1] = stv_paramDefine( "Volume Rate", (void *)&vrate,
			stvFloat, stvVisOnly ); 

/* --- Declare a decomposition of density */
	/* -- it is block, block, collapsed */

	axisType[0] = stvBlock;
	axisType[1] = stvBlock;
	axisType[2] = stvCollapse;

	prank = 2;
	pshape[0] = ROWS;
	pshape[1] = COLUMNS;

	decompId[0] = stv_decompDefine( 3, axisType,
			stvDefaultInfo, stvDefaultInfo, glb, gub, prank, pshape );

/* --- declare the density data field with the above decomposition */

	paddr[0] = myrow;
	paddr[1] = mycol;

	declare[0] = IDIM;
	declare[1] = JDIM;
	declare[2] = KDIM;

	fieldId[0] = stv_fieldDefine( density, "density", decompId[0],
			stvNoFieldOffsets, declare, stvFloat, paddr, stvVisOnly );

	/* set simulation time step for density field to 1 msec */

	stv_setopt( fieldId[0], stvOptFieldTimeScale,
			stv_time_of( 1, stvTimeMilliSec ) );

	/* set base time to 5 mins + 40 secs + 10 msecs = 340,010 msecs */

	stv_setopt( fieldId[0], stvOptSetFieldTime,
			stv_time_of( 5, stvTimeMin ) );

	stv_setopt( fieldId[0], stvOptIncrFieldTime,
			stv_time_of( 40, stvTimeSec ) );

	stv_setopt( fieldId[0], stvOptIncrFieldTime,
			stv_time_of( 10, stvTimeMilliSec ) );

/* --- Verify Local Field Decomposition */

	stv_dumpLocalFieldBounds( fieldId[0], stvVerbose );

	stv_dumpLocalFieldBounds( fieldId[0], stvTerse );

/* ---  Top of main computational loop */

	for( ; ; )
	{
		radius = radius + rate;
		sequence = sequence + 1;
		vradius = sequence*vrate;

		if ( myinum == 0 ) 
			printf( "Sequence %d\n", sequence );

		for ( k=1 ; k <= KDIM ; k++ )	
		{

#ifdef INTERMEDIATE_SEND
			/*  --- send intermediate data to any attached viewers */
			stv_sendReadyData( stvSendNoFieldTimeUpdates );
#endif

			globalk = k;

			for ( j=1 ; j <= JDIM ; j++ )
			{
				/*	Determine from the center of the compgrid, which 
					points should Should be colored */

				globalj = j + mycol * JDIM;

				for( i=1 ; i <= IDIM ; i++ )
				{
					globali = i + myrow * IDIM;

					dist = SQR( globali - centeri )
							+ SQR( globalk - centerk );

					dist = sqrt( dist + SQR( globalj - centerj ) );

					if ( dist < radius ) 
						value = ( 1.0 - dist / radius) * rate * vradius;

					else
						value = 0.0;

					density[ i - 1 ][ j - 1 ][ k - 1 ] = value;
				}
			}
		}

/*  --- send a message to mod(myinum+1, ntasks).
			Keep the tasks pseudo-synched */

		pvm_initsend( PvmDataDefault );
		pvm_pkint( &myinum, 1, 1 );
		pvm_pkint( &mytid, 1, 1 );

		pvm_send( dest, RINGMSG );

		bufid = pvm_recv( -1, RINGMSG );
		pvm_bufinfo(bufid, &cc, &cc, &msrc);

 		if ( msrc != src )
		{
			printf(
				"Ring message seq %d from %x instead of %x, exiting.\n",
					sequence, msrc, src);
			pvm_bcast( CSIMPLEGRP, RINGMSG);
			pvm_exit();
			exit(-1);
		}

		pvm_upkint( &onum, 1, 1 );
		pvm_upkint( &otid, 1, 1 );

/*  --- send messages to any attached CUMULVS front-end viewers */

		info = stv_sendReadyData( stvSendDefault );

		if ( info > 0 && myinum == 0 ) 
			dump_params_changed( info, paramIds, rate, vrate );
	}
}


/* print out message that parameters have been updated by a viewer */

void
dump_params_changed( info, paramIds, rate, vrate )
int info;
int *paramIds;
float rate;
float vrate;
{
	int i;

	printf( "stv_sendReadyData returned %d\n", info );

	for ( i=0 ; i < NPARAMS ; i++ )
	{
		if( stv_isParamChanged( paramIds[i] ) )
		{
		 	if ( i == 0 )
				printf( "Parameter 'Radius Rate' was changed to %f\n",
						rate );
			else if ( i == 1 )	 	
				printf( "Parameter 'Volume Rate' was changed to %f\n",
						vrate );
		}
	}
}

